==> gnome-chemistry-utils-0.4.8.tar.bz2: gnome-chemistry-utils-0.4.8/gcu/crystalview.cc <==
/* 
 * Gnome Chemisty Utils
 * crystaldoc.cc 
 *
 * Copyright (C) 2002-2005
 *
 * Developed by Jean Bréfort <jean.brefort@normalesup.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

#include "config.h"
#include "crystalview.h"
#include "crystaldoc.h"
#include "matrix.h"
#include "xml-utils.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <gtk/gtkgl.h>
#include <math.h>

static GdkGLConfig *glconfig = NULL;

using namespace gcu;

typedef struct
{
	unsigned glList;
} WidgetData;

static bool on_init(GtkWidget *widget, CrystalView* View) 
{
	View->Init(widget);
	return true;
}

bool on_reshape(GtkWidget *widget, GdkEventConfigure *event, CrystalView* View) 
{
	View->Reshape(widget);
	return true;
}

static bool on_draw(GtkWidget *widget, GdkEventExpose *event, CrystalView* View) 
{
	/* Draw only last expose. */
	if (event->count > 0) return TRUE;

	View->Draw(widget);
	return true;
}

static bool on_motion(GtkWidget *widget, GdkEventMotion *event, CrystalView* View) 
{
	View->OnMotion(widget, event);
	return true;
}

static bool on_pressed(GtkWidget *widget, GdkEventButton *event, CrystalView* View) 
{
	return View->OnPressed(widget, event);
}

static void on_destroyed(GtkWidget *widget, CrystalView *pView)
{
	pView->OnDestroyed(widget);
}

CrystalView::CrystalView(CrystalDoc* pDoc)
{
	m_pDoc = pDoc;
	m_bInit = false;
}

CrystalView::~CrystalView()
{
}

void CrystalView::Init(GtkWidget *widget)
{
	GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
	GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
	if (gdk_gl_drawable_gl_begin(gldrawable, glcontext))
	{
	    glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CULL_FACE);
		glEnable(GL_COLOR_MATERIAL);
		float shiny = 25.0, spec[4] = {1.0, 1.0, 1.0, 1.0};
		glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, &shiny);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, spec);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		m_bInit = true;
		Update(widget);
    }
}

bool CrystalView::Load (xmlNodePtr node)
{
	char *txt;
	xmlNodePtr child = node->children;
	while (child) {
		if (!strcmp ((gchar*) child->name, "orientation")) {
			txt = (char*) xmlGetProp (child, (xmlChar*) "psi");
			if (txt) {
				sscanf(txt, "%lg", &m_psi);
				xmlFree (txt);
			}	else
				return false;
			txt = (char*) xmlGetProp (child, (xmlChar*) "theta");
			if (txt) {
				sscanf(txt, "%lg", &m_theta);
				xmlFree (txt);
			}	else
				return false;
			txt = (char*) xmlGetProp (child, (xmlChar*) "phi");
			if (txt) {
				sscanf(txt, "%lg", &m_phi); 
				xmlFree (txt);
			}	else
				return false;
			Matrix m (m_psi / 180 * M_PI, m_theta / 180 * M_PI, m_phi / 180 * M_PI, euler);
			m_Euler = m;
		} else if (!strcmp ((gchar*) child->name, "fov")) {
			txt = (char*) xmlNodeGetContent (child);
			int result = sscanf (txt, "%lg", &m_fAngle);
			if (result == 0)
				m_fAngle = 10.;
			xmlFree (txt);
		}
		child = child->next;
	}
	if (!ReadColor (node, "background", &m_fRed, &m_fGreen, &m_fBlue, &m_fAlpha))
		return false;
	return true;
}

xmlNodePtr CrystalView::Save(xmlDocPtr xml)
{
	xmlNodePtr parent, child;
	gchar buf[256];
	parent = xmlNewDocNode(xml, NULL, (xmlChar*)"view", NULL);
	if (!parent) return NULL;
	
	child = xmlNewDocNode(xml, NULL, (xmlChar*)"orientation", NULL);
	if (child) xmlAddChild(parent, child);
	else {xmlFreeNode(parent); return NULL;}
	snprintf(buf, sizeof(buf), "%g", m_psi);
	xmlNewProp(child, (xmlChar*)"psi", (xmlChar*)buf);
	snprintf(buf, sizeof(buf), "%g", m_theta);
	xmlNewProp(child, (xmlChar*)"theta", (xmlChar*)buf);
	snprintf(buf, sizeof(buf), "%g", m_phi);
	xmlNewProp(child, (xmlChar*)"phi", (xmlChar*)buf);
	
	g_snprintf(buf, sizeof(buf) - 1, "%g", m_fAngle);
	child = xmlNewDocNode(xml, NULL, (xmlChar*)"fov", (xmlChar*)buf);
	if (child) xmlAddChild(parent, child);
	else {xmlFreeNode(parent); return NULL;}
	
	if (!WriteColor(xml, parent, "background", m_fRed, m_fGreen, m_fBlue, m_fAlpha)) {xmlFreeNode(parent); return NULL;}
	
	return parent;
}

GtkWidget* CrystalView::CreateNewWidget()
{
/* Create new OpenGL widget. */
	if (glconfig == NULL)
	{
		/* Check if OpenGL is supported. */
		if (!gdk_gl_query_extension())
		{
			g_print("\n*** OpenGL is not supported.\n");
			exit(1);
		}
	
		/* Configure OpenGL-capable visual. */
	
		/* Try double-buffered visual */
		glconfig = gdk_gl_config_new_by_mode(GdkGLConfigMode(GDK_GL_MODE_RGB |
											GDK_GL_MODE_DEPTH |
											GDK_GL_MODE_DOUBLE));
		if (glconfig == NULL)
		{
			g_print("*** Cannot find the double-buffered visual.\n");
			  exit(1);
		}
	}
	/* create new OpenGL widget */
	m_pWidget = GTK_WIDGET(gtk_drawing_area_new());

	/* Set OpenGL-capability to the widget. */
	gtk_widget_set_gl_capability(m_pWidget,
                                glconfig,
                                NULL,
                                TRUE,
                                GDK_GL_RGBA_TYPE);

	WidgetData* pData = new WidgetData;
	pData->glList = 0;
	g_object_set_data(G_OBJECT(m_pWidget), "gldata", pData);
	m_Widgets.push_back(m_pWidget);
// Events for widget must be set before X Window is created
	gtk_widget_set_events(GTK_WIDGET(m_pWidget),
			GDK_EXPOSURE_MASK|
			GDK_BUTTON_MOTION_MASK|
			GDK_POINTER_MOTION_HINT_MASK|
			GDK_BUTTON_PRESS_MASK);

// Connect signal handlers
// Do initialization when widget has been realized.
	g_signal_connect(G_OBJECT(m_pWidget), "realize",
		     GTK_SIGNAL_FUNC(on_init), this);
// When window is resized viewport needs to be resized also.
	g_signal_connect(G_OBJECT(m_pWidget), "configure_event",
		     GTK_SIGNAL_FUNC(on_reshape), this);
// Redraw image when exposed. 
	g_signal_connect(G_OBJECT(m_pWidget), "expose_event",
		     GTK_SIGNAL_FUNC(on_draw), this);
// When moving mouse 
  g_signal_connect (G_OBJECT(m_pWidget), "motion_notify_event",
		      GTK_SIGNAL_FUNC(on_motion), this);
// When a mouse button is pressed
  g_signal_connect (G_OBJECT(m_pWidget), "button_press_event",
		      GTK_SIGNAL_FUNC(on_pressed), this);
// When a widget is destroyed
	g_signal_connect (G_OBJECT(m_pWidget), "destroy", GTK_SIGNAL_FUNC(on_destroyed), this);

	gtk_widget_show(GTK_WIDGET(m_pWidget));
	return m_pWidget;
}

bool CrystalView::OnPressed(GtkWidget *widget, GdkEventButton *event)
{
  if (event->button == 1) {
    // beginning of drag, reset mouse position
    m_lastx = event->x;
    m_lasty = event->y;
    return true;
  }
  return false;
}

void CrystalView::OnMotion(GtkWidget *widget, GdkEventMotion *event)
{
	gint x, y;
	GdkModifierType state;
	
	if (event->is_hint)
		gdk_window_get_pointer(event->window, &x, &y, &state);
	else
	{
	    x = (gint) event->x;
	    y = (gint) event->y;
	    state = (GdkModifierType) event->state;
	}
	if (state & GDK_BUTTON1_MASK)
	{
		if ((x == m_lastx) && (y == m_lasty)) return;
			m_pDoc->SetDirty();
		Rotate(x - m_lastx, y - m_lasty);
		m_lastx = x;
		m_lasty = y;
		std::list<GtkWidget*>::iterator i;
		for (i = m_Widgets.begin(); i!= m_Widgets.end(); i++)
		{
			gtk_widget_queue_draw_area(*i, 0, 0, (*i)->allocation.width, (*i)->allocation.height);
		}
	}
}

void CrystalView::Reshape(GtkWidget *widget)
{
	if (!m_bInit) return;
	float fAspect;
	GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
	GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
	if (gdk_gl_drawable_gl_begin(gldrawable, glcontext))
    {
		if (widget->allocation.height)
		{
			fAspect = (GLfloat)widget->allocation.width/(GLfloat)widget->allocation.height;
			if (fAspect == 0.0) fAspect = 1.0;
		}
		else	// don't divide by zero, not that we should ever run into that...
			fAspect = 1.0f;
		double x = m_pDoc->GetMaxDist();
		if (x == 0) x = 1;
		m_fRadius = (float) (x / sin(m_fAngle / 360 * M_PI)) ;
		glViewport(0,0, widget->allocation.width, widget->allocation.height);
		if (fAspect > 1.0)
		{
			m_height = x * (1 - tan(m_fAngle / 360 * M_PI));
			m_width = m_height * fAspect;
		}
		else
		{
			m_width = x * (1 - tan(m_fAngle / 360 * M_PI));
			m_height = m_width / fAspect;
		}
		m_near = m_fRadius - x;
		m_far = m_fRadius + x;
	    glMatrixMode(GL_PROJECTION);
	    glLoadIdentity();
		glFrustum(- m_width, m_width, - m_height, m_height, m_near , m_far);
	    glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslatef(0, 0, -m_fRadius);
	}
}

void CrystalView::Draw(GtkWidget *widget)
{
	if (!m_bInit) return;
	GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
	GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
	if (gdk_gl_drawable_gl_begin(gldrawable, glcontext))
    {
		glClearColor(m_fRed, m_fGreen, m_fBlue, m_fAlpha);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		WidgetData* pData = (WidgetData*)g_object_get_data(G_OBJECT(widget), "gldata");
		m_nGLList = pData->glList;
		if  (m_nGLList)
		{
			glPushMatrix();
			glRotated(m_psi, 0.0, 1.0, 0.0);
			glRotated(m_theta, 0.0, 0.0, 1.0);
			glRotated(m_phi, 0.0, 1.0, 0.0);
			glEnable (GL_BLEND);
			glCallList(m_nGLList);
			glDisable (GL_BLEND);
			glPopMatrix();
		}
	/* Swap backbuffer to front */
		gdk_gl_drawable_swap_buffers(gldrawable);
    }
}

void CrystalView::Update()
{
	std::list<GtkWidget*>::iterator i;
	for (i = m_Widgets.begin(); i!= m_Widgets.end(); i++) Update(*i);
}

void CrystalView::Update(GtkWidget* widget)
{
	if (!m_bInit) return;
	Reshape(widget);
	GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
	GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
	if (gdk_gl_drawable_gl_begin(gldrawable, glcontext))
 	{
		WidgetData* pData = (WidgetData*)g_object_get_data(G_OBJECT(widget), "gldata");
		m_nGLList = pData->glList;
		if (m_nGLList) glDeleteLists(m_nGLList,1);
		pData->glList = m_nGLList = glGenLists(1);
		glNewList(m_nGLList, GL_COMPILE);
		m_pDoc->Draw();
		glEndList();
	}
	Draw(widget);
}

void CrystalView::OnDestroyed (GtkWidget *widget)
{
	delete (WidgetData*) g_object_get_data (G_OBJECT (widget), "gldata");
	m_Widgets.remove (widget);
}

void CrystalView::Rotate(gdouble x, gdouble y)
{
	gdouble z = sqrt(x*x + y*y);
	Matrix Mat(0, (y > 0) ? - acos(x/z) : acos(x/z), z * 0.00349065850398866, rotation);
	m_Euler = Mat * m_Euler;
	m_Euler.Euler(m_psi, m_theta, m_phi);
	m_psi /= 0.0174532925199433;
	m_theta /= 0.0174532925199433;
	m_phi /= 0.0174532925199433;
}

==> gnome-chemistry-utils-0.7.1.tar.bz2: gnome-chemistry-utils-0.7.1/gcu/crystalview.cc <==
/* 
 * Gnome Chemisty Utils
 * crystalview.cc 
 *
 * Copyright (C) 2002-2005 Jean Bréfort <jean.brefort@normalesup.org>
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

#include "config.h"
#include "crystalview.h"
#include "crystaldoc.h"
#include "matrix.h"
#include "xml-utils.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <gtk/gtkgl.h>
#include <math.h>
#include <glib/gi18n-lib.h>
#include <libgnomevfs/gnome-vfs-ops.h>

static GdkGLConfig *glconfig = NULL;

using namespace gcu;

typedef struct
{
	unsigned glList;
} WidgetData;

static bool on_init(GtkWidget *widget, CrystalView* View) 
{
	View->Init(widget);
	return true;
}

bool on_reshape(GtkWidget *widget, GdkEventConfigure *event, CrystalView* View) 
{
	View->Reshape(widget);
	return true;
}

static bool on_draw(GtkWidget *widget, GdkEventExpose *event, CrystalView* View) 
{
	/* Draw only last expose. */
	if (event->count > 0) return TRUE;

	View->Draw(widget);
	return true;
}

static bool on_motion(GtkWidget *widget, GdkEventMotion *event, CrystalView* View) 
{
	View->OnMotion(widget, event);
	return true;
}

static bool on_pressed(GtkWidget *widget, GdkEventButton *event, CrystalView* View) 
{
	return View->OnPressed(widget, event);
}

static void on_destroyed(GtkWidget *widget, CrystalView *pView)
{
	pView->OnDestroyed(widget);
}

CrystalView::CrystalView(CrystalDoc* pDoc)
{
	m_pDoc = pDoc;
	m_bInit = false;
}

CrystalView::~CrystalView()
{
}

void CrystalView::Init(GtkWidget *widget)
{
	GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
	GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
	if (gdk_gl_drawable_gl_begin(gldrawable, glcontext))
	{
	    glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CULL_FACE);
		glEnable(GL_COLOR_MATERIAL);
		float shiny = 25.0, spec[4] = {1.0, 1.0, 1.0, 1.0};
		glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, &shiny);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, spec);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		m_bInit = true;
		gdk_gl_drawable_gl_end (gldrawable);
		Update(widget);
    }
}

bool CrystalView::Load (xmlNodePtr node)
{
	char *txt;
	xmlNodePtr child = node->children;
	while (child) {
		if (!strcmp ((gchar*) child->name, "orientation")) {
			txt = (char*) xmlGetProp (child, (xmlChar*) "psi");
			if (txt) {
				sscanf(txt, "%lg", &m_psi);
				xmlFree (txt);
			}	else
				return false;
			txt = (char*) xmlGetProp (child, (xmlChar*) "theta");
			if (txt) {
				sscanf(txt, "%lg", &m_theta);
				xmlFree (txt);
			}	else
				return false;
			txt = (char*) xmlGetProp (child, (xmlChar*) "phi");
			if (txt) {
				sscanf(txt, "%lg", &m_phi); 
				xmlFree (txt);
			}	else
				return false;
			Matrix m (m_psi / 180 * M_PI, m_theta / 180 * M_PI, m_phi / 180 * M_PI, euler);
			m_Euler = m;
		} else if (!strcmp ((gchar*) child->name, "fov")) {
			txt = (char*) xmlNodeGetContent (child);
			int result = sscanf (txt, "%lg", &m_fAngle);
			if (result == 0)
				m_fAngle = 10.;
			xmlFree (txt);
		}
		child = child->next;
	}
	if (!ReadColor (node, "background", &m_fRed, &m_fGreen, &m_fBlue, &m_fAlpha))
		return false;
	return true;
}

xmlNodePtr CrystalView::Save(xmlDocPtr xml)
{
	xmlNodePtr parent, child;
	gchar buf[256];
	parent = xmlNewDocNode(xml, NULL, (xmlChar*)"view", NULL);
	if (!parent) return NULL;
	
	child = xmlNewDocNode(xml, NULL, (xmlChar*)"orientation", NULL);
	if (child) xmlAddChild(parent, child);
	else {xmlFreeNode(parent); return NULL;}
	snprintf(buf, sizeof(buf), "%g", m_psi);
	xmlNewProp(child, (xmlChar*)"psi", (xmlChar*)buf);
	snprintf(buf, sizeof(buf), "%g", m_theta);
	xmlNewProp(child, (xmlChar*)"theta", (xmlChar*)buf);
	snprintf(buf, sizeof(buf), "%g", m_phi);
	xmlNewProp(child, (xmlChar*)"phi", (xmlChar*)buf);
	
	g_snprintf(buf, sizeof(buf) - 1, "%g", m_fAngle);
	child = xmlNewDocNode(xml, NULL, (xmlChar*)"fov", (xmlChar*)buf);
	if (child) xmlAddChild(parent, child);
	else {xmlFreeNode(parent); return NULL;}
	
	if (!WriteColor(xml, parent, "background", m_fRed, m_fGreen, m_fBlue, m_fAlpha)) {xmlFreeNode(parent); return NULL;}
	
	return parent;
}

GtkWidget* CrystalView::CreateNewWidget()
{
/* Create new OpenGL widget. */
	if (glconfig == NULL)
	{
		/* Check if OpenGL is supported. */
		if (!gdk_gl_query_extension())
		{
			g_print("\n*** OpenGL is not supported.\n");
			exit(1);
		}
	
		/* Configure OpenGL-capable visual. */
	
		/* Try double-buffered visual */
		glconfig = gdk_gl_config_new_by_mode(GdkGLConfigMode(GDK_GL_MODE_RGB |
											GDK_GL_MODE_DEPTH |
											GDK_GL_MODE_DOUBLE));
		if (glconfig == NULL)
		{
			g_print("*** Cannot find the double-buffered visual.\n");
			  exit(1);
		}
	}
	/* create new OpenGL widget */
	m_pWidget = GTK_WIDGET(gtk_drawing_area_new());

	/* Set OpenGL-capability to the widget. */
	gtk_widget_set_gl_capability(m_pWidget,
                                glconfig,
                                NULL,
                                TRUE,
                                GDK_GL_RGBA_TYPE);

	WidgetData* pData = new WidgetData;
	pData->glList = 0;
	g_object_set_data(G_OBJECT(m_pWidget), "gldata", pData);
	m_Widgets.push_back(m_pWidget);
// Events for widget must be set before X Window is created
	gtk_widget_set_events(GTK_WIDGET(m_pWidget),
			GDK_EXPOSURE_MASK|
			GDK_BUTTON_MOTION_MASK|
			GDK_POINTER_MOTION_HINT_MASK|
			GDK_BUTTON_PRESS_MASK);

// Connect signal handlers
// Do initialization when widget has been realized.
	g_signal_connect(G_OBJECT(m_pWidget), "realize",
		     GTK_SIGNAL_FUNC(on_init), this);
// When window is resized viewport needs to be resized also.
	g_signal_connect(G_OBJECT(m_pWidget), "configure_event",
		     GTK_SIGNAL_FUNC(on_reshape), this);
// Redraw image when exposed. 
	g_signal_connect(G_OBJECT(m_pWidget), "expose_event",
		     GTK_SIGNAL_FUNC(on_draw), this);
// When moving mouse 
  g_signal_connect (G_OBJECT(m_pWidget), "motion_notify_event",
		      GTK_SIGNAL_FUNC(on_motion), this);
// When a mouse button is pressed
  g_signal_connect (G_OBJECT(m_pWidget), "button_press_event",
		      GTK_SIGNAL_FUNC(on_pressed), this);
// When a widget is destroyed
	g_signal_connect (G_OBJECT(m_pWidget), "destroy", GTK_SIGNAL_FUNC(on_destroyed), this);

	gtk_widget_show(GTK_WIDGET(m_pWidget));
	return m_pWidget;
}

bool CrystalView::OnPressed(GtkWidget *widget, GdkEventButton *event)
{
  if (event->button == 1) {
    // beginning of drag, reset mouse position
    m_lastx = event->x;
    m_lasty = event->y;
    return true;
  }
  return false;
}

void CrystalView::OnMotion(GtkWidget *widget, GdkEventMotion *event)
{
	gint x, y;
	GdkModifierType state;
	
	if (event->is_hint)
		gdk_window_get_pointer(event->window, &x, &y, &state);
	else
	{
	    x = (gint) event->x;
	    y = (gint) event->y;
	    state = (GdkModifierType) event->state;
	}
	if (state & GDK_BUTTON1_MASK)
	{
		if ((x == m_lastx) && (y == m_lasty)) return;
			m_pDoc->SetDirty();
		Rotate(x - m_lastx, y - m_lasty);
		m_lastx = x;
		m_lasty = y;
		std::list<GtkWidget*>::iterator i;
		for (i = m_Widgets.begin(); i!= m_Widgets.end(); i++)
		{
			gtk_widget_queue_draw_area(*i, 0, 0, (*i)->allocation.width, (*i)->allocation.height);
		}
	}
}

void CrystalView::Reshape(GtkWidget *widget)
{
	if (!m_bInit) return;
	float fAspect;
	GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
	GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
	if (gdk_gl_drawable_gl_begin(gldrawable, glcontext))
    {
		if (widget->allocation.height)
		{
			fAspect = (GLfloat)widget->allocation.width/(GLfloat)widget->allocation.height;
			if (fAspect == 0.0) fAspect = 1.0;
		}
		else	// don't divide by zero, not that we should ever run into that...
			fAspect = 1.0f;
		double x = m_pDoc->GetMaxDist();
		if (x == 0) x = 1;
		m_fRadius = (float) (x / sin(m_fAngle / 360 * M_PI)) ;
		glViewport(0,0, widget->allocation.width, widget->allocation.height);
		if (fAspect > 1.0)
		{
			m_height = x * (1 - tan(m_fAngle / 360 * M_PI));
			m_width = m_height * fAspect;
		}
		else
		{
			m_width = x * (1 - tan(m_fAngle / 360 * M_PI));
			m_height = m_width / fAspect;
		}
		m_near = m_fRadius - x;
		m_far = m_fRadius + x;
	    glMatrixMode(GL_PROJECTION);
	    glLoadIdentity();
		glFrustum(- m_width, m_width, - m_height, m_height, m_near , m_far);
	    glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslatef(0, 0, -m_fRadius);
		gdk_gl_drawable_gl_end (gldrawable);
	}
}

void CrystalView::Draw(GtkWidget *widget)
{
	if (!m_bInit) return;
	GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
	GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
	if (gdk_gl_drawable_gl_begin(gldrawable, glcontext))
    {
		glClearColor(m_fRed, m_fGreen, m_fBlue, m_fAlpha);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		WidgetData* pData = (WidgetData*)g_object_get_data(G_OBJECT(widget), "gldata");
		m_nGLList = pData->glList;
		if  (m_nGLList)
		{
			glPushMatrix();
			glRotated(m_psi, 0.0, 1.0, 0.0);
			glRotated(m_theta, 0.0, 0.0, 1.0);
			glRotated(m_phi, 0.0, 1.0, 0.0);
			glEnable (GL_BLEND);
			glCallList(m_nGLList);
			glDisable (GL_BLEND);
			glPopMatrix();
		}
		gdk_gl_drawable_gl_end (gldrawable);
	/* Swap backbuffer to front */
		gdk_gl_drawable_swap_buffers(gldrawable);
    }
}

void CrystalView::Update()
{
	std::list<GtkWidget*>::iterator i;
	for (i = m_Widgets.begin(); i!= m_Widgets.end(); i++) Update(*i);
}

void CrystalView::Update(GtkWidget* widget)
{
	if (!m_bInit) return;
	Reshape(widget);
	GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
	GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
	if (gdk_gl_drawable_gl_begin(gldrawable, glcontext))
 	{
		WidgetData* pData = (WidgetData*)g_object_get_data(G_OBJECT(widget), "gldata");
		m_nGLList = pData->glList;
		if (m_nGLList) glDeleteLists(m_nGLList,1);
		pData->glList = m_nGLList = glGenLists(1);
		glNewList(m_nGLList, GL_COMPILE);
		m_pDoc->Draw();
		glEndList();
		gdk_gl_drawable_gl_end (gldrawable);
	}
	Draw(widget);
}

void CrystalView::OnDestroyed (GtkWidget *widget)
{
	delete (WidgetData*) g_object_get_data (G_OBJECT (widget), "gldata");
	m_Widgets.remove (widget);
}

void CrystalView::Rotate(gdouble x, gdouble y)
{
	gdouble z = sqrt(x*x + y*y);
	Matrix Mat(0, (y > 0) ? - acos(x/z) : acos(x/z), z * 0.00349065850398866, rotation);
	m_Euler = Mat * m_Euler;
	m_Euler.Euler(m_psi, m_theta, m_phi);
	m_psi /= 0.0174532925199433;
	m_theta /= 0.0174532925199433;
	m_phi /= 0.0174532925199433;
}

static gboolean do_save_image (const gchar *buf, gsize count, GError **error, gpointer data)
{
	GnomeVFSHandle *handle = (GnomeVFSHandle*) data;
	gsize written = 0;
	GnomeVFSResult res;
	while (count) {
		res = gnome_vfs_write (handle, buf, count, &written);
		if (res != GNOME_VFS_OK) {
			g_set_error (error, g_quark_from_static_string ("gchemutils"), res, gnome_vfs_result_to_string (res));
			return false;
		}
		count -= written;
	}
	return true;
}

void CrystalView::SaveAsImage (char const *filename, char const *type, map<string, string>& options)
{
	int w = m_pWidget->allocation.width;
	int h = m_pWidget->allocation.height;
	GdkGLConfig *glconfig = gdk_gl_config_new_by_mode (
		GdkGLConfigMode (GDK_GL_MODE_RGB | GDK_GL_MODE_DEPTH));
	GdkPixmap *pixmap = gdk_pixmap_new (
			(GdkDrawable*) (m_pWidget->window),
			w, h, -1);
	GdkGLPixmap *gl_pixmap = gdk_pixmap_set_gl_capability (pixmap,
							       glconfig,
							       NULL );
	GdkGLDrawable * drawable = gdk_pixmap_get_gl_drawable (pixmap);
	GdkGLContext * context = gdk_gl_context_new (drawable,
						     NULL,
						     FALSE,
						     GDK_GL_RGBA_TYPE);
	if (gdk_gl_drawable_gl_begin (drawable, context)) {
	    glEnable (GL_LIGHTING);
		glEnable (GL_LIGHT0);
		glEnable (GL_DEPTH_TEST);
		glEnable (GL_CULL_FACE);
		glEnable (GL_COLOR_MATERIAL);
		float shiny = 25.0, spec[4] = {1.0, 1.0, 1.0, 1.0};
		glMaterialfv (GL_FRONT_AND_BACK, GL_SHININESS, &shiny);
		glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, spec);
		glViewport (0, 0, w, h);
	    glMatrixMode (GL_PROJECTION);
	    glLoadIdentity ();
		glFrustum (- m_width, m_width, - m_height, m_height, m_near , m_far);
	    glMatrixMode (GL_MODELVIEW);
		glLoadIdentity ();
		glTranslatef (0, 0, -m_fRadius);
		glClearColor (m_fRed, m_fGreen, m_fBlue, m_fAlpha);
		glClearDepth (1.0);
		glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		WidgetData* pData = (WidgetData*) g_object_get_data (G_OBJECT (m_pWidget), "gldata");
		m_nGLList = pData->glList;
		if  (m_nGLList) {
			glPushMatrix ();
			glRotated (m_psi, 0.0, 1.0, 0.0);
			glRotated (m_theta, 0.0, 0.0, 1.0);
			glRotated (m_phi, 0.0, 1.0, 0.0);
			glEnable (GL_BLEND);
			m_pDoc->Draw();
			glDisable (GL_BLEND);
			glPopMatrix ();
		}
		glFlush ();
		gdk_gl_drawable_gl_end (drawable);
		GdkPixbuf* pixbuf = gdk_pixbuf_get_from_drawable (NULL,
			(GdkDrawable*) pixmap, NULL, 0, 0, 0, 0, -1, -1);
		char const **keys = g_new0 (char const*, options.size () + 1);
		char const **values = g_new0 (char const*, options.size ());
		GError *error = NULL;
		map<string, string>::iterator i, iend = options.end ();
		int j = 0;
		for (i = options.begin (); i != iend; i++) {
			keys[j] = (*i).first.c_str ();
			values[j++] = (*i).second.c_str ();
		}
		GnomeVFSHandle *handle = NULL;
		if (gnome_vfs_create (&handle, filename, GNOME_VFS_OPEN_WRITE, true, 0644) == GNOME_VFS_OK) {
			gdk_pixbuf_save_to_callbackv (pixbuf, do_save_image, handle, type, (char**) keys, (char**) values, &error);
			if (error) {
				fprintf (stderr, _("Unable to save image file: %s\n"), error->message);
				g_error_free (error);
			}
			gnome_vfs_close (handle); // hope there will be no error there
		}
		g_free (keys);
		g_free (values);
		g_object_unref (pixbuf);
	}

	gdk_gl_context_destroy (context);
	gdk_gl_pixmap_destroy (gl_pixmap);
	// destroying pixmap gives a CRITICAL and destroying glconfig leeds to a crash.
}
